@@ -5,9 +5,9 @@
 ;; Author: Daniel Mendler <mail@daniel-mendler.de>
 ;; Maintainer: Daniel Mendler <mail@daniel-mendler.de>
 ;; Created: 2022
-;; Version: 1.2
-;; Package-Requires: ((emacs "27.1") (compat "30"))
-;; Homepage: https://github.com/minad/tempel
+;; Version: 1.3
+;; Package-Requires: ((emacs "28.1") (compat "30"))
+;; URL: https://github.com/minad/tempel
 ;; Keywords: abbrev, languages, tools, text
 
 ;; This file is part of GNU Emacs.
@@ -48,7 +48,7 @@
 (defgroup tempel nil
   "Tempo templates/snippets with in-buffer field editing."
   :link '(info-link :tag "Info Manual" "(tempel)")
-  :link '(url-link :tag "Homepage" "https://github.com/minad/tempel")
+  :link '(url-link :tag "Website" "https://github.com/minad/tempel")
   :link '(emacs-library-link :tag "Library Source" "tempel.el")
   :group 'abbrev
   :group 'tools
@@ -76,11 +76,11 @@ trigger completion."
 
 (defcustom tempel-insert-annotation 40
   "Annotation width for `tempel-insert'."
-  :type '(choice (const nil natnum)))
+  :type '(choice (const nil) natnum))
 
 (defcustom tempel-complete-annotation 20
   "Annotation width for `tempel-complete'."
-  :type '(choice (const nil natnum)))
+  :type '(choice (const nil) natnum))
 
 (defcustom tempel-user-elements nil
   "List of user element handler functions.
@@ -133,6 +133,9 @@ If a file is modified, added or removed, reload the templates."
 (defvar tempel--inhibit-hooks nil
   "Inhibit tempel modification change hooks from running.")
 
+(defvar tempel--ignore-condition nil
+  "Ignore template condition.")
+
 (defvar-local tempel--active nil
   "List of active templates.
 Each template state is a pair, where the car is a list of overlays and
@@ -172,16 +175,17 @@ may be named with `tempel--name' or carry an evaluatable Lisp expression
      ((or 'n '& '% 'o) #("\n" 0 1 (face completions-annotations)))
      (_ #("_" 0 1 (face shadow))))))
 
-(defun tempel--annotate (templates width ellipsis sep name)
+(defun tempel--annotate (templates width sep name)
   "Annotate template NAME given the list of TEMPLATES.
-WIDTH, SEP and ELLIPSIS configure the formatting."
+WIDTH and SEP configure the formatting."
   (when-let ((name (intern-soft name))
              (elts (cdr (assoc name templates))))
-    (concat sep (truncate-string-to-width
-                 (replace-regexp-in-string
-                  "[ \t\n\r]+" #(" " 0 1 (face completions-annotations))
-                  (tempel--print-template elts))
-                 width 0 ?\s ellipsis))))
+    (concat sep (string-trim
+                 (truncate-string-to-width
+                  (replace-regexp-in-string
+                   "[ \t\n\r]+" #(" " 0 1 (face completions-annotations))
+                   (tempel--print-template elts))
+                  width)))))
 
 (defun tempel--info-buffer (templates fun name)
   "Create info buffer for template NAME.
@@ -254,9 +258,13 @@ BEG and END are the boundaries of the modification."
         (save-excursion
           (goto-char (overlay-start ov))
           (let (x)
-            (setq x (or (and (setq x (overlay-get ov 'tempel--form)) (or (eval x (cdr st)) ""))
-                        (and (setq x (overlay-get ov 'tempel--name)) (alist-get x (cdr st)))))
-            (when x (tempel--synchronize-replace (overlay-start ov) (overlay-end ov) ov x)))))
+            (setq x (or (and (setq x (overlay-get ov 'tempel--form))
+                             (or (eval x (cdr st)) ""))
+                        (and (setq x (overlay-get ov 'tempel--name))
+                             (alist-get x (cdr st)))))
+            (when x
+              (tempel--synchronize-replace (overlay-start ov)
+                                           (overlay-end ov) ov x)))))
       ;; Move range overlay
       (move-overlay range (overlay-start range)
                     (max (overlay-end range) (overlay-end ov))))))
@@ -365,7 +373,13 @@ Return the added field."
          (indent-region (car region) (cdr region) nil))))
     ;; TEMPEL EXTENSION: Quit template immediately
     ('q (overlay-put (tempel--field st) 'tempel--enter #'tempel--done))
-    (_ (if-let ((ret (run-hook-with-args-until-success 'tempel-user-elements elt)))
+    (_ (if-let ((ret (run-hook-wrapped 'tempel-user-elements
+                                       (lambda (hook elt fields)
+                                         (condition-case nil
+                                             (funcall hook elt)
+                                           (wrong-number-of-arguments
+                                            (funcall hook elt fields))))
+                                       elt (cdr st))))
            (tempel--element st region ret)
          ;; TEMPEL EXTENSION: Evaluate forms
          (tempel--form st elt)))))
@@ -476,10 +490,9 @@ This is meant to be a source in `tempel-template-sources'."
            (timestamps
             (cl-loop
              for f in files collect
-             (cons f (time-convert
-                      (file-attribute-modification-time
-                       (file-attributes (file-truename f)))
-                      'integer)))))
+             (cons f (time-convert (file-attribute-modification-time
+                                    (file-attributes (file-truename f)))
+                                   'integer)))))
       (unless (equal (car tempel--path-templates) timestamps)
         (setq tempel--path-templates (cons timestamps
                                            (mapcan #'tempel--file-read files))))))
@@ -497,7 +510,8 @@ This is meant to be a source in `tempel-template-sources'."
         (derived-mode-p m)
         (when-let ((remap (alist-get m (bound-and-true-p major-mode-remap-alist))))
           (derived-mode-p remap))))
-   (or (not (plist-member plist :when))
+   (or tempel--ignore-condition
+       (not (plist-member plist :when))
        (save-excursion
          (save-restriction
            (save-match-data
@@ -510,8 +524,8 @@ This is meant to be a source in `tempel-template-sources'."
      'tempel-template-sources
      (lambda (fun)
        (cond
-        ((functionp fun) (setq result (append result (funcall fun))))
-        ((boundp fun) (setq result (append result (symbol-value fun))))
+        ((functionp fun) (cl-callf append result (funcall fun)))
+        ((boundp fun) (cl-callf append result (symbol-value fun)))
         (t (error "Template source is not a function or a variable: %S" fun)))
        nil))
     result))
@@ -537,14 +551,20 @@ This is meant to be a source in `tempel-template-sources'."
            ((and (< dir 0) (< stop pt))
             (setq next (max (or next -1) stop)))))))))
 
+(defun tempel--active-p (_sym buffer)
+  "Return non-nil if Tempel is active in BUFFER."
+  (buffer-local-value 'tempel--active buffer))
+
 (defun tempel-beginning ()
   "Move to beginning of the template."
+  (declare (completion tempel--active-p))
   (interactive)
   (when-let ((pos (tempel--beginning)))
     (if (= pos (point)) (tempel-done) (goto-char pos))))
 
 (defun tempel-end ()
   "Move to end of the template."
+  (declare (completion tempel--active-p))
   (interactive)
   (when-let ((pos (tempel--end)))
     (if (= pos (point)) (tempel-done) (goto-char pos))))
@@ -560,6 +580,7 @@ This is meant to be a source in `tempel-template-sources'."
 
 (defun tempel-kill ()
   "Kill the field contents."
+  (declare (completion tempel--active-p))
   (interactive)
   (if-let ((ov (tempel--field-at-point)))
       (kill-region (overlay-start ov) (overlay-end ov))
@@ -567,6 +588,7 @@ This is meant to be a source in `tempel-template-sources'."
 
 (defun tempel-next (arg)
   "Move ARG fields forward and quit at the end."
+  (declare (completion tempel--active-p))
   (interactive "p")
   (cl-loop for i below (abs arg) do
            (if-let ((next (tempel--find arg)))
@@ -580,6 +602,7 @@ This is meant to be a source in `tempel-template-sources'."
 
 (defun tempel-previous (arg)
   "Move ARG fields backward and quit at the beginning."
+  (declare (completion tempel--active-p))
   (interactive "p")
   (tempel-next (- arg)))
 
@@ -595,6 +618,7 @@ This is meant to be a source in `tempel-template-sources'."
 
 (defun tempel-abort ()
   "Abort template insertion."
+  (declare (completion tempel--active-p))
   (interactive)
   ;; TODO abort only the topmost template?
   (while-let ((st (car tempel--active)))
@@ -606,7 +630,7 @@ This is meant to be a source in `tempel-template-sources'."
 (defun tempel--disable (&optional st)
   "Disable template ST, or last template."
   (if st
-      (setq tempel--active (delq st tempel--active))
+      (cl-callf2 delq st tempel--active)
     (setq st (pop tempel--active)))
   (when st
     (mapc #'delete-overlay (car st))
@@ -615,6 +639,7 @@ This is meant to be a source in `tempel-template-sources'."
 
 (defun tempel-done ()
   "Template completion is done."
+  (declare (completion tempel--active-p))
   (interactive)
   ;; TODO disable only the topmost template?
   (while tempel--active (tempel--done)))
@@ -636,14 +661,6 @@ This is meant to be a source in `tempel-template-sources'."
     (tempel--save)
     (or (completion-at-point) (user-error "%s: No completions" capf))))
 
-(defun tempel--completion-table (templates)
-  "Return a completion table for a list of TEMPLATES.
-The completion table specifies the category `tempel'."
-  (lambda (str pred action)
-    (if (eq action 'metadata)
-        '(metadata (category . tempel))
-      (complete-with-action action templates str pred))))
-
 (defun tempel--prefix-bounds ()
   "Return prefix bounds."
   (if tempel-trigger-prefix
@@ -676,8 +693,8 @@ command."
                (sym (intern-soft name))
                (template (assq sym templates)))
       (setq templates (list template))
-      (list (car bounds) (cdr bounds)
-            (tempel--completion-table templates)
+      (list (car bounds) (cdr bounds) templates
+            :category 'tempel
             :exclusive 'no
             :exit-function (apply-partially #'tempel--exit templates nil)))))
 
@@ -699,8 +716,8 @@ Capf, otherwise like an interactive completion command."
       (when-let ((templates (tempel--templates))
                  (bounds (or (and (not region) (tempel--prefix-bounds))
                              (and (not tempel-trigger-prefix) (cons (point) (point))))))
-        (list (car bounds) (cdr bounds)
-              (tempel--completion-table templates)
+        (list (car bounds) (cdr bounds) templates
+              :category 'tempel
               :exclusive 'no
               :company-kind (lambda (_) 'snippet)
               :exit-function (apply-partially #'tempel--exit templates region)
@@ -718,7 +735,7 @@ Capf, otherwise like an interactive completion command."
               :annotation-function
               (and tempel-complete-annotation
                    (apply-partially #'tempel--annotate
-                                    templates tempel-complete-annotation nil " ")))))))
+                                    templates tempel-complete-annotation " ")))))))
 
 ;;;###autoload
 (defun tempel-insert (template-or-name)
@@ -727,18 +744,24 @@ If called interactively, select a template with `completing-read'."
   (interactive (list nil))
   (tempel--insert
    (if (consp template-or-name) template-or-name
-     (let* ((templates (or (tempel--templates)
-                           (error "Tempel: No templates for %s" major-mode)))
-            (completion-extra-properties
-             (and tempel-insert-annotation
-                  (list :annotation-function
-                        (apply-partially
-                         #'tempel--annotate templates tempel-insert-annotation t
-                         #("  " 1 2 (display (space :align-to (+ left 20)))))))))
+     (let ((templates (or (tempel--templates)
+                          (error "Tempel: No templates for %s" major-mode))))
        (unless template-or-name
-         (setq template-or-name (intern-soft
-                                 (completing-read "Template: " templates
-                                                  nil t nil 'tempel--history))))
+         (setq template-or-name
+               (intern-soft
+                (completing-read
+                 "Template: "
+                 (lambda (str pred action)
+                   (if (eq action 'metadata)
+                       `(metadata
+                         (category . tempel)
+                         ,@(when tempel-insert-annotation
+                             `((annotation-function
+                                . ,(apply-partially
+                                    #'tempel--annotate templates tempel-insert-annotation
+                                    #("  " 1 2 (display (space :align-to (+ left 20)))))))))
+                     (complete-with-action action templates str pred)))
+                 nil t nil 'tempel--history))))
        (or (and template-or-name (alist-get template-or-name templates))
            (user-error "Template %s not found" template-or-name))))
    (tempel--region)))
@@ -761,12 +784,6 @@ If called interactively, select a template with `completing-read'."
                (interactive)
                (tempel-insert ',template-or-name)))))))
 
-(defun tempel--abbrev-hook (name template)
-  "Abbreviation expansion hook for TEMPLATE with NAME."
-  (tempel--delete-word name)
-  (tempel--insert template nil)
-  t)
-
 ;;;###autoload
 (define-minor-mode tempel-abbrev-mode
   "Install Tempel templates as abbrevs."
@@ -778,13 +795,19 @@ If called interactively, select a template with `completing-read'."
             (default-value 'abbrev-minor-mode-table-alist))
     (kill-local-variable 'abbrev-minor-mode-table-alist))
   (when tempel-abbrev-mode
-    (let ((table (make-abbrev-table)))
+    (let ((table (make-abbrev-table))
+          (tempel--ignore-condition t))
       (dolist (template (tempel--templates))
         (let* ((name (symbol-name (car template)))
                (hook (make-symbol name)))
-          (fset hook (apply-partially #'tempel--abbrev-hook name (cdr template)))
+          (fset hook (lambda ()
+                       (tempel--delete-word name)
+                       (tempel--insert (cdr template) nil)
+                       t))
           (put hook 'no-self-insert t)
-          (define-abbrev table name 'Template hook :system t)))
+          (define-abbrev table name 'Template hook
+            :system t :enable-function
+            (lambda () (assq (car template) (tempel--templates))))))
       (setq-local abbrev-minor-mode-table-alist
                   (cons `(tempel-abbrev-mode . ,table)
                         abbrev-minor-mode-table-alist)))))
@@ -799,14 +822,5 @@ If called interactively, select a template with `completing-read'."
   (unless (or noninteractive (eq (aref (buffer-name) 0) ?\s))
     (tempel-abbrev-mode 1)))
 
-;; Emacs 28: Do not show Tempel commands in M-X
-(dolist (sym (list #'tempel-next #'tempel-previous #'tempel-beginning
-                   #'tempel-end #'tempel-kill #'tempel-done #'tempel-abort))
-  (put sym 'completion-predicate #'tempel--command-p))
-
-(defun tempel--command-p (_sym buffer)
-  "Return non-nil if Tempel is active in BUFFER."
-  (buffer-local-value 'tempel--active buffer))
-
 (provide 'tempel)
 ;;; tempel.el ends here
